(* ========================================================================= *)
(*                                                                           *)
(*               Stability_Resonance.ml                       *)
(*                                                                           *)
(*        (c) Copyright, Sidi Mohamed BEILLAHI, Muhammad Umair Siddique      *)
(*                       2015                                                *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact:	<beillahi@encs.concordia.ca>		       		     *)
(*  	        <muh_sidd@ece.concordia.ca>                                  *)
(*       	                                    			     *)
(*                                                                           *)
(* Last update: March 02, 2015                                               *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)



needs "Mason_Gain.ml";;


(*------------------------------------------------------*)
(*----------------Formalization-------------------------*)
(*------------------------------------------------------*)
(*--------------------***************-------------------*)
(*--Denominator and numerator of the transfer function--*)

let DENOMINATOR = define 
   `(DENOMINATOR (sfg : SFG) = 
    (DETERMINANT (EC (fst_of_four sfg) (snd_of_four sfg))))` ;;
		
let NUMERATOR_AID = define `!(sfg: SFG) (P: (num) list) (H: ((num) list) list) 
	(G: ((num) list) list).
	NUMERATOR_AID sfg P H G =  PRODUCT_FORWARD_DELTA (FC_REWRITING 
	(FORWARD_CIRCUIT 0 (EC_replace_vector_head P (lst_of_four sfg)) 
	  H G (thd_of_four sfg) (snd_of_four sfg)) (fst_of_four sfg))
	  (EC_REWRITING (EC_ALL_POINT 0 (EC_replace_vector_head P 1) 
	   H G (snd_of_four sfg) (snd_of_four sfg) P) (fst_of_four sfg))`;;

let NUMERATOR = define `!(sfg: SFG).
	NUMERATOR sfg
	= if((fst_of_four sfg)=[]) then Cx(&0) else  
	(NUMERATOR_AID sfg (EC_create_vector (snd_of_four sfg)) 
	(EC_create_matrix (snd_of_four sfg)) 
	(EC_G_creation 0 (EC_start_nodes (fst_of_four sfg)) 
	(EC_end_nodes (fst_of_four sfg)) (EC_create_vector (snd_of_four sfg)))) ` ;;

(*------------------------------------------------------------------*)
let SFG_Theorem = 
	prove( `!(sfg:SFG). SFG_Main sfg = (NUMERATOR sfg)/(DENOMINATOR sfg) `,
	REWRITE_TAC [SFG_Main;NUMERATOR;DENOMINATOR;MASON_GAIN]);;

	
(*------------------Stability definitions---------------------*)
let set_of_poles = 
	define `!(sfg: real^2->SFG) 
	(z:complex). set_of_poles  sfg z =  
	{z | DENOMINATOR (sfg z) = Cx(&0)}`;; 
	
let is_valid_pole = 
	define `!(sfg: complex->SFG) 
	(z:complex) (p:complex). 
	is_valid_pole sfg z p <=> p IN  (set_of_poles sfg z)`;; 
	
let are_valid_poles = 
	define `!(sfg: complex->SFG) 
	(z:complex) (t:(complex) list). 
	(are_valid_poles sfg z [] = T) /\ 
	(are_valid_poles sfg z (CONS h t) = 
	((is_valid_pole sfg z h) /\ (are_valid_poles sfg z t)))`;;

let in_unit_circle = 
	define `!(p:complex). in_unit_circle p = norm p <= &1`;; 
	
let poles_in_unit_circle = 
	define `!(t:(complex) list). (poles_in_unit_circle [] = T) /\ 
	(poles_in_unit_circle (CONS h t) = 
	( (in_unit_circle h) /\ poles_in_unit_circle t))`;;


let is_stable = 
	define `!(sfg: complex->SFG) 
	(z:complex) (PL:(complex) list). 
	is_stable sfg z PL <=> 
	poles_in_unit_circle PL /\ are_valid_poles sfg z PL`;;

let  unstable  = 
	define `!(sfg: complex->SFG) 
	(z:complex) (PL:(complex) list).
	unstable sfg z PL = ~(is_stable sfg z PL)`;;

(*---------------Resonance Definitions-------------------*)

let set_of_zeros = 
	define `!(sfg: real^2->SFG) 
	(z:complex). set_of_zeros  sfg z =  
	{z | NUMERATOR (sfg z) = Cx(&0)}`;; 
	
let is_valid_zero = 
	define `!(sfg: complex->SFG) 
	(z:complex) (p:complex). 
	is_valid_zero sfg z p <=> p IN  (set_of_zeros sfg z)`;; 
	
let are_valid_zeros = 
	define `!(sfg: complex->SFG) 
	(z:complex) (t:(complex) list). 
	(are_valid_zeros sfg z [] = T) /\ 
	(are_valid_zeros sfg z (CONS h t) = 
	((is_valid_zero sfg z h) /\ (are_valid_zeros sfg z t)))`;;

let is_resonant = 
	define `!(sfg: complex->SFG) 
	(z:complex) (PL:(complex) list). 
	is_resonant sfg z PL <=> 
	poles_in_unit_circle PL /\ are_valid_zeros sfg z PL`;;

let  inresonant  = 
	define `!(sfg: complex->SFG) 
	(z:complex) (PL:(complex) list).
	inresonant sfg z PL = ~(is_resonant sfg z PL)`;;
(*-----------------------------------------------------------------------*)
(* 			Finite and cardinality theorem                   *)
(*-----------------------------------------------------------------------*)
let CARD_SET_POLES_THM = prove (`!n c. ~(!i. i IN 0..n ==> c i = Cx (&0)) /\ 
	    (!z. DENOMINATOR (sfg z) =  vsum (0..n) (\i. c i * z pow i) ) ==>  
	     FINITE (set_of_poles sfg z) /\   CARD (set_of_poles sfg z) <= n`, 
		  SIMP_TAC[COMPLEX_POLYFUN_ROOTBOUND;set_of_poles]);;
		  
let CARD_SET_ZEROS_THM = prove (`!n c. ~(!i. i IN 0..n ==> c i = Cx (&0)) /\ 
	    (!z. NUMERATOR (sfg z) =  vsum (0..n) (\i. c i * z pow i) ) ==>  
	    FINITE (set_of_zeros sfg z) /\   CARD (set_of_zeros sfg z) <= n`, 
	      SIMP_TAC[COMPLEX_POLYFUN_ROOTBOUND;set_of_zeros]);;



(*-----------------------------------------------------------------------*)

let NUMERATOR_SIMP_TAC = 
 REWRITE_TAC [NUMERATOR;component_of_four;NOT_CONS_NIL]
THEN NUMBER_SFG_TAC1 [minus; EC_H_MODIFIE1; 
LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE; EC_H_MODIFIE0; 
EC_remove_node;EC_add_node;  
EC_TEST; IN_LIST; EECTEST;   
EC_add_one_to_list_head; EC_create_matrix; EC_replace_vector_head; 
 EC_create_vector; EC_G_creation;  
 EC_start_nodes ;  
EC_G_creation_aid; EC_end_nodes;HD; 
TL;component_of_four;EL;APPEND]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus;HD;TL;component_of_four;EL;APPEND;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus; EC_H_MODIFIE1;LENGTH; 
EC_H_MODIFIE2;EC_H_MODIFIE; EC_H_MODIFIE0; EC_remove_node; 
EC_add_node;EC_TEST; IN_LIST; EECTEST;   
EC_add_one_to_list_head; EC_create_matrix; EC_replace_vector_head; 
 EC_create_vector; EC_G_creation;  
 EC_start_nodes ;  
EC_G_creation_aid; EC_end_nodes;HD; 
TL;component_of_trpl;EL;APPEND;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN NUMBER_SFG_TAC1 [NUMERATOR_AID;
EC_add_one_to_list_head;HD;component_of_four] THEN  
NUMBER_SFG_TAC2 [NUMERATOR_AID;HD;component_of_four]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus] THEN  
NUMBER_SFG_TAC2 [minus_one_aid]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [NUMERATOR_AID;EL;minus;HD;TL;EECTEST;FST; 
SND;EC_replace_vector_head;APPEND;component_of_four] THEN  
NUMBER_SFG_TAC2 [NUMERATOR_AID;EL;minus;HD;TL;EECTEST;FST; 
SND;EC_replace_vector_head;APPEND;component_of_four] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN ONCE_REWRITE_TAC [ EC_ALL_POINT]  
THEN REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;ECSKIP] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node;IN_LIST] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node;IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;;



let DENOMINATOR_SIMP_TAC2 =
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EECTEST;component_of_trpl;APPEND;EC_add_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EECTEST;component_of_trpl;APPEND;EC_add_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EECTEST;component_of_trpl;APPEND;EC_add_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EECTEST;component_of_trpl;APPEND;EC_add_node] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN NUMBER_SFG_TAC1 [ EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND] THEN CONV_TAC NUM_REDUCE_CONV THEN   
REWRITE_TAC [NOT_CONS_NIL] THEN 
NUMBER_SFG_TAC1 [EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND;REPLICATE] THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [ EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND;REPLICATE] THEN REWRITE_TAC [NOT_CONS_NIL] THEN 
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD] THEN  
NUMBER_SFG_TAC2 [HD]  THEN REWRITE_TAC [NOT_CONS_NIL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN
(REPEAT (CHANGED_TAC MASON_SIMP_TAC));;


let DENOMINATOR_SIMP_TAC1 =
REWRITE_TAC [DENOMINATOR; component_of_four] THEN 
REWRITE_TAC [EC;NOT_CONS_NIL] 
THEN NUMBER_SFG_TAC1 [ EC_MAIN; minus;  
EC_H_MODIFIE1;LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE; EC_H_MODIFIE0;  
EC_remove_node;EC_add_node;  
 EC_TEST; IN_LIST; EECTEST;   
EC_add_one_to_list_head; EC_create_matrix;  
EC_replace_vector_head;  
EC_create_vector; EC_G_creation;  
EC_start_nodes ;  
EC_G_creation_aid; EC_end_nodes;HD;TL;
component_of_trpl;EL;APPEND;REPLICATE]  
 THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [ minus;HD;TL;component_of_trpl;EL; 
APPEND;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [ EC_MAIN; minus; EC_H_MODIFIE1; 
LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE; EC_H_MODIFIE0;  
EC_remove_node;EC_add_node;  
 EC_TEST; IN_LIST; EECTEST;   
EC_add_one_to_list_head; EC_create_matrix;  
EC_replace_vector_head;  
EC_create_vector; EC_G_creation;  
EC_start_nodes ;  
EC_G_creation_aid; EC_end_nodes;HD;TL;component_of_trpl
;EL;APPEND;REPLICATE] THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EC_add_one_to_list_head;HD;REPLICATE] THEN  
NUMBER_SFG_TAC2 [HD;REPLICATE]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus] THEN  
NUMBER_SFG_TAC2 [minus_one_aid]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EECTEST;component_of_trpl;APPEND] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EECTEST;component_of_trpl;APPEND] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN ONCE_REWRITE_TAC [ EC_ALL_POINT] 
THEN REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;ECSKIP] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node;IN_LIST] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node;IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;;



let NUMERATOR_TAC = NUMERATOR_SIMP_TAC
THEN (REPEAT (CHANGED_TAC TAC_EC))
THEN ECFEC_SIMP_TAC2
THEN (REPEAT (CHANGED_TAC TAC_FC))
THEN ECFEC_SIMP_TAC3 ;;

let DENOMINATOR_TAC = DENOMINATOR_SIMP_TAC1
THEN (REPEAT (CHANGED_TAC TAC_EC))
THEN DENOMINATOR_SIMP_TAC2;;
